@@ -13,7 +13,7 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-from __future__ import with_statement
+import glob
 import hashlib
 import os
 import sqlite3
@@ -21,14 +21,18 @@ import threading
 import time
 
 from mapproxy.image import ImageSource
-from mapproxy.cache.base import TileCacheBase, tile_buffer, CacheBackendError
+from mapproxy.cache.base import TileCacheBase, tile_buffer, REMOVE_ON_UNLOCK
 from mapproxy.util.fs import ensure_directory
 from mapproxy.util.lock import FileLock
-from mapproxy.compat import BytesIO, PY2
+from mapproxy.compat import BytesIO, PY2, itertools
 
 import logging
 log = logging.getLogger(__name__)
 
+if not hasattr(glob, 'escape'):
+    import re
+    glob.escape = lambda pathname: re.sub(r'([*?[])', r'[\1]', pathname)
+
 def sqlite_datetime_to_timestamp(datetime):
     if datetime is None:
         return None
@@ -38,10 +42,12 @@ def sqlite_datetime_to_timestamp(datetime):
 class MBTilesCache(TileCacheBase):
     supports_timestamp = False
 
-    def __init__(self, mbtile_file, with_timestamps=False):
+    def __init__(self, mbtile_file, with_timestamps=False, timeout=30, wal=False):
         self.lock_cache_id = 'mbtiles-' + hashlib.md5(mbtile_file.encode('utf-8')).hexdigest()
         self.mbtile_file = mbtile_file
         self.supports_timestamp = with_timestamps
+        self.timeout = timeout
+        self.wal = wal
         self.ensure_mbtile()
         self._db_conn_cache = threading.local()
 
@@ -49,7 +55,7 @@ class MBTilesCache(TileCacheBase):
     def db(self):
         if not getattr(self._db_conn_cache, 'db', None):
             self.ensure_mbtile()
-            self._db_conn_cache.db = sqlite3.connect(self.mbtile_file)
+            self._db_conn_cache.db = sqlite3.connect(self.mbtile_file, self.timeout)
         return self._db_conn_cache.db
 
     def cleanup(self):
@@ -62,8 +68,8 @@ class MBTilesCache(TileCacheBase):
 
     def ensure_mbtile(self):
         if not os.path.exists(self.mbtile_file):
-            with FileLock(os.path.join(os.path.dirname(self.mbtile_file), 'init.lck'),
-                remove_on_unlock=True):
+            with FileLock(self.mbtile_file + '.init.lck',
+                remove_on_unlock=REMOVE_ON_UNLOCK):
                 if not os.path.exists(self.mbtile_file):
                     ensure_directory(self.mbtile_file)
                     self._initialize_mbtile()
@@ -71,6 +77,10 @@ class MBTilesCache(TileCacheBase):
     def _initialize_mbtile(self):
         log.info('initializing MBTile file %s', self.mbtile_file)
         db  = sqlite3.connect(self.mbtile_file)
+
+        if self.wal:
+            db.execute('PRAGMA journal_mode=wal')
+
         stmt = """
             CREATE TABLE tiles (
                 zoom_level integer,
@@ -135,25 +145,42 @@ class MBTilesCache(TileCacheBase):
     def store_tile(self, tile):
         if tile.stored:
             return True
-        with tile_buffer(tile) as buf:
-            if PY2:
-                content = buffer(buf.read())
-            else:
-                content = buf.read()
-            x, y, level = tile.coord
-            cursor = self.db.cursor()
-            try:
-                if self.supports_timestamp:
-                    stmt = "INSERT OR REPLACE INTO tiles (zoom_level, tile_column, tile_row, tile_data, last_modified) VALUES (?,?,?,?, datetime(?, 'unixepoch', 'localtime'))"
-                    cursor.execute(stmt, (level, x, y, content, time.time()))
+        return self._store_bulk([tile])
+
+    def store_tiles(self, tiles):
+        tiles = [t for t in tiles if not t.stored]
+        return self._store_bulk(tiles)
+
+    def _store_bulk(self, tiles):
+        records = []
+        # tile_buffer (as_buffer) will encode the tile to the target format
+        # we collect all tiles before, to avoid having the db transaction
+        # open during this slow encoding
+        for tile in tiles:
+            with tile_buffer(tile) as buf:
+                if PY2:
+                    content = buffer(buf.read())
                 else:
-                    stmt = "INSERT OR REPLACE INTO tiles (zoom_level, tile_column, tile_row, tile_data) VALUES (?,?,?,?)"
-                    cursor.execute(stmt, (level, x, y, content))
-                self.db.commit()
-            except sqlite3.OperationalError as ex:
-                log.warn('unable to store tile: %s', ex)
-                return False
-            return True
+                    content = buf.read()
+                x, y, level = tile.coord
+                if self.supports_timestamp:
+                    records.append((level, x, y, content, time.time()))
+                else:
+                    records.append((level, x, y, content))
+
+        cursor = self.db.cursor()
+        try:
+            if self.supports_timestamp:
+                stmt = "INSERT OR REPLACE INTO tiles (zoom_level, tile_column, tile_row, tile_data, last_modified) VALUES (?,?,?,?, datetime(?, 'unixepoch', 'localtime'))"
+                cursor.executemany(stmt, records)
+            else:
+                stmt = "INSERT OR REPLACE INTO tiles (zoom_level, tile_column, tile_row, tile_data) VALUES (?,?,?,?)"
+                cursor.executemany(stmt, records)
+            self.db.commit()
+        except sqlite3.OperationalError as ex:
+            log.warning('unable to store tile: %s', ex)
+            return False
+        return True
 
     def load_tile(self, tile, with_metadata=False):
         if tile.source or tile.coord is None:
@@ -198,29 +225,35 @@ class MBTilesCache(TileCacheBase):
             # all tiles loaded or coords are None
             return True
 
-        if len(coords) > 1000:
-            # SQLite is limited to 1000 args
-            raise CacheBackendError('cannot query SQLite for more than 333 tiles')
-
         if self.supports_timestamp:
-            stmt = "SELECT tile_column, tile_row, tile_data, last_modified FROM tiles WHERE "
+            stmt_base = "SELECT tile_column, tile_row, tile_data, last_modified FROM tiles WHERE "
         else:
-            stmt = "SELECT tile_column, tile_row, tile_data FROM tiles WHERE "
-        stmt += ' OR '.join(['(tile_column = ? AND tile_row = ? AND zoom_level = ?)'] * (len(coords)//3))
-
-        cursor = self.db.cursor()
-        cursor.execute(stmt, coords)
+            stmt_base = "SELECT tile_column, tile_row, tile_data FROM tiles WHERE "
 
         loaded_tiles = 0
-        for row in cursor:
-            loaded_tiles += 1
-            tile = tile_dict[(row[0], row[1])]
-            data = row[2]
-            tile.size = len(data)
-            tile.source = ImageSource(BytesIO(data))
-            if self.supports_timestamp:
-                tile.timestamp = sqlite_datetime_to_timestamp(row[3])
-        cursor.close()
+
+        # SQLite is limited to 1000 args -> split into multiple requests if more arguments are needed
+        while coords:
+            cur_coords = coords[:999]
+
+            stmt = stmt_base + ' OR '.join(
+                ['(tile_column = ? AND tile_row = ? AND zoom_level = ?)'] * (len(cur_coords) // 3))
+
+            cursor = self.db.cursor()
+            cursor.execute(stmt, cur_coords)
+
+            for row in cursor:
+                loaded_tiles += 1
+                tile = tile_dict[(row[0], row[1])]
+                data = row[2]
+                tile.size = len(data)
+                tile.source = ImageSource(BytesIO(data))
+                if self.supports_timestamp:
+                    tile.timestamp = sqlite_datetime_to_timestamp(row[3])
+            cursor.close()
+
+            coords = coords[999:]
+
         return loaded_tiles == len(tile_dict)
 
     def remove_tile(self, tile):
@@ -265,10 +298,12 @@ class MBTilesCache(TileCacheBase):
 class MBTilesLevelCache(TileCacheBase):
     supports_timestamp = True
 
-    def __init__(self, mbtiles_dir):
+    def __init__(self, mbtiles_dir, timeout=30, wal=False):
         self.lock_cache_id = 'sqlite-' + hashlib.md5(mbtiles_dir.encode('utf-8')).hexdigest()
         self.cache_dir = mbtiles_dir
         self._mbtiles = {}
+        self.timeout = timeout
+        self.wal = wal
         self._mbtiles_lock = threading.Lock()
 
     def _get_level(self, level):
@@ -281,6 +316,8 @@ class MBTilesLevelCache(TileCacheBase):
                 self._mbtiles[level] = MBTilesCache(
                     mbtile_filename,
                     with_timestamps=True,
+                    timeout=self.timeout,
+                    wal=self.wal,
                 )
 
         return self._mbtiles[level]
@@ -307,6 +344,14 @@ class MBTilesLevelCache(TileCacheBase):
 
         return self._get_level(tile.coord[2]).store_tile(tile)
 
+    def store_tiles(self, tiles):
+        failed = False
+        for level, tiles in itertools.groupby(tiles, key=lambda t: t.coord[2]):
+            tiles = [t for t in tiles if not t.stored]
+            res = self._get_level(level).store_tiles(tiles)
+            if not res: failed = True
+        return failed
+
     def load_tile(self, tile, with_metadata=False):
         if tile.source or tile.coord is None:
             return True
@@ -340,7 +385,8 @@ class MBTilesLevelCache(TileCacheBase):
         if timestamp == 0:
             level_cache.cleanup()
             os.unlink(level_cache.mbtile_file)
+            for file in glob.glob("%s-*" % glob.escape(level_cache.mbtile_file)):
+                os.unlink(file)
             return True
         else:
             return level_cache.remove_level_tiles_before(level, timestamp)
-
