@@ -36,7 +36,7 @@ def validate_options(conf_dict):
     else:
         return [], True
 
-coverage = {
+coverage = recursive({
     'polygons': str(),
     'polygons_srs': str(),
     'bbox': one_of(str(), [number()]),
@@ -47,7 +47,13 @@ coverage = {
     'datasource': one_of(str(), [number()]),
     'where': str(),
     'srs': str(),
-}
+    'expire_tiles': str(),
+    'union': [recursive()],
+    'difference': [recursive()],
+    'intersection': [recursive()],
+    'clip': bool(),
+})
+
 image_opts = {
     'mode': str(),
     'colors': number(),
@@ -65,6 +71,7 @@ http_opts = {
     'client_timeout': number(),
     'ssl_no_cert_checks': bool(),
     'ssl_ca_certs': str(),
+    'hide_error_details': bool(),
     'headers': {
         anything(): str()
     },
@@ -106,12 +113,23 @@ cache_types = {
     },
     'sqlite': {
         'directory': str(),
+        'sqlite_timeout': number(),
+        'sqlite_wal': bool(),
         'tile_lock_dir': str(),
     },
     'mbtiles': {
         'filename': str(),
+        'sqlite_timeout': number(),
+        'sqlite_wal': bool(),
         'tile_lock_dir': str(),
     },
+    'geopackage': {
+        'filename': str(),
+        'directory': str(),
+        'tile_lock_dir': str(),
+        'table_name': str(),
+        'levels': bool(),
+    },
     'couchdb': {
         'url': str(),
         'db_name': str(),
@@ -121,6 +139,16 @@ cache_types = {
         'tile_id': str(),
         'tile_lock_dir': str(),
     },
+    's3': {
+        'bucket_name': str(),
+        'directory_layout': str(),
+        'directory': str(),
+        'profile_name': str(),
+        'region_name': str(),
+        'endpoint_url': str(),
+        'access_control_list': str(),
+        'tile_lock_dir': str(),
+     },
     'riak': {
         'nodes': [riak_node],
         'protocol': one_of('pbc', 'http', 'https'),
@@ -130,7 +158,20 @@ cache_types = {
             'http': number(),
         },
         'secondary_index': bool(),
-    }
+        'tile_lock_dir': str(),
+    },
+    'redis': {
+        'host': str(),
+        'port': int(),
+        'db': int(),
+        'prefix': str(),
+        'default_ttl': int(),
+    },
+    'compact': {
+        'directory': str(),
+        required('version'): number(),
+        'tile_lock_dir': str(),
+    },
 }
 
 on_error = {
@@ -283,6 +324,20 @@ ogc_service_md = {
     ],
 }
 
+band_source = {
+    required('source'): str(),
+    required('band'): int,
+    'factor': number(),
+}
+
+band_sources = {
+    'r': [band_source],
+    'g': [band_source],
+    'b': [band_source],
+    'a': [band_source],
+    'l': [band_source],
+}
+
 mapproxy_yaml_spec = {
     '__config_files__': anything(), # only used internaly
     'globals': {
@@ -310,10 +365,17 @@ mapproxy_yaml_spec = {
             'tile_lock_dir': str(),
             'meta_size': [number()],
             'meta_buffer': number(),
+            'bulk_meta_tiles': bool(),
             'max_tile_limit': number(),
             'minimize_meta_requests': bool(),
             'concurrent_tile_creators': int(),
             'link_single_color_images': bool(),
+            's3': {
+                'bucket_name': str(),
+                'profile_name': str(),
+                'region_name': str(),
+                'endpoint_url': str(),
+            },
         },
         'grid': {
             'tile_size': [int()],
@@ -322,6 +384,7 @@ mapproxy_yaml_spec = {
           'axis_order_ne': [str()],
           'axis_order_en': [str()],
           'proj_data_dir': str(),
+          'preferred_src_proj': {anything(): [str()]},
         },
         'tiles': {
             'expires_hours': number(),
@@ -336,12 +399,13 @@ mapproxy_yaml_spec = {
     },
     'caches': {
         anything(): {
-            required('sources'): [string_type],
+            required('sources'): one_of([string_type], band_sources),
             'name': str(),
             'grids': [str()],
             'cache_dir': str(),
             'meta_size': [number()],
             'meta_buffer': number(),
+            'bulk_meta_tiles': bool(),
             'minimize_meta_requests': bool(),
             'concurrent_tile_creators': int(),
             'disable_storage': bool(),
@@ -351,6 +415,9 @@ mapproxy_yaml_spec = {
             'use_direct_from_level': number(),
             'use_direct_from_res': number(),
             'link_single_color_images': bool(),
+            'cache_rescaled_tiles': bool(),
+            'upscale_tiles': int(),
+            'downscale_tiles': int(),
             'watermark': {
                 'text': string_type,
                 'font_size': number(),
@@ -374,7 +441,14 @@ mapproxy_yaml_spec = {
             'kvp': bool(),
             'restful': bool(),
             'restful_template': str(),
+            'restful_featureinfo_template': str(),
             'md': ogc_service_md,
+            'featureinfo_formats': [
+                {
+                    required('mimetype'): str(),
+                    'suffix': str(),
+                },
+            ],
         },
         'wms': {
             'srs': [str()],
@@ -407,6 +481,7 @@ mapproxy_yaml_spec = {
                     'legendurl': str(),
                     'featureinfo_format': str(),
                     'featureinfo_xslt': str(),
+                    'featureinfo_out_format': str(),
                 },
                 'image': combined(image_opts, {
                     'opacity':number(),
@@ -416,6 +491,7 @@ mapproxy_yaml_spec = {
                 'supported_formats': [str()],
                 'supported_srs': [str()],
                 'http': http_opts,
+                'on_error': on_error,
                 'forward_req_params': [str()],
                 required('req'): {
                     required('url'): str(),
@@ -464,6 +540,23 @@ mapproxy_yaml_spec = {
                 'use_mapnik2': bool(),
                 'scale_factor': number(),
             }),
+            'arcgis': combined(source_commons, {
+               required('req'): {
+                    required('url'): str(),
+                    'dpi': int(),
+                    'layers': str(),
+                    'transparent': bool(),
+                    'time': str()
+                },
+                'opts': {
+                    'featureinfo': bool(),
+                    'featureinfo_tolerance': number(),
+                    'featureinfo_return_geometries': bool(),
+                },
+                'supported_srs': [str()],
+                'http': http_opts,
+                'on_error': on_error
+            }),
             'debug': {
             },
         })
@@ -480,6 +573,7 @@ mapproxy_yaml_spec = {
         },
         recursive([combined(scale_hints, {
             'sources': [string_type],
+            'tile_sources': [string_type],
             'name': str(),
             required('title'): string_type,
             'legendurl': str(),
@@ -498,13 +592,3 @@ mapproxy_yaml_spec = {
     'parts': anything(),
 }
 
-if __name__ == '__main__':
-    import sys
-    import yaml
-    for f in sys.argv[1:]:
-        data = yaml.load(open(f))
-        try:
-            validate(mapproxy_yaml_spec, data)
-        except ValidationError as ex:
-            for err in ex.errors:
-                print('%s: %s' % (f, err))
